/******************************************************************************
* choria
* Copyright (c) 2023 Alan Witkowski
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/
#pragma once

// Libraries
#include <ae/type.h>
#include <objects/action.h>
#include <constants.h>
#include <unordered_map>
#include <vector>
#include <list>
#include <cstdint>

// Forward Declarations
class _Object;
class _Stats;
class _Battle;
class _HUD;
class _StatusEffect;
struct _Vendor;
struct _Trader;
struct _MinigameType;
struct _Blacksmith;
struct _Enchanter;

// Structures
struct _Unlock {
	int Level{0};
};

struct _SetData {
	int EquippedCount{0};
	int MaxLevel{0};
	int Level{0};
};

struct _Cooldown {
	double Duration{0.0};
	double MaxDuration{0.0};
};

// Classes
class _Character {

	public:

		enum StatusImageType {
			STATUS_NONE,
			STATUS_MENU,
			STATUS_INVENTORY,
			STATUS_VENDOR,
			STATUS_SKILLS,
			STATUS_TRADE,
			STATUS_TRADER,
			STATUS_BLACKSMITH,
			STATUS_MINIGAME,
			STATUS_BATTLE,
			STATUS_TELEPORT,
			STATUS_DEAD,
		};

		_Character(_Object *Object);
		~_Character();

		// Initialize
		void Init();

		// Updates
		void Update(double FrameTime);
		void UpdateHealth(int64_t &Value);
		void UpdateMana(int64_t Value);
		void UpdateGold(int64_t Value);
		void UpdateExperience(int64_t Value);
		void UpdateAllResist(int Value);
		void UpdateElementalResist(int Value);

		// Stats
		void CalculateStats();
		void CalculateLevelStats();
		float GetNextLevelPercent() const;
		bool IsAlive() const { return Attributes.at("Health").Int64 > 0; }
		float GetHealthPercent() const { return Attributes.at("MaxHealth").Int64 > 0 ? Attributes.at("Health").Int64 / (float)Attributes.at("MaxHealth").Int64 : 0; }
		float GetManaPercent() const { return Attributes.at("MaxMana").Int64 > 0 ? Attributes.at("Mana").Int64 / (float)Attributes.at("MaxMana").Int64 : 0; }
		int64_t GetItemCost(int64_t ItemCost);

		// Input
		bool AcceptingMoveInput();
		bool CanOpenTrade() const { return IsAlive() && !Battle; }
		bool CanOpenSkills() const { return IsAlive() && !Battle; }
		bool CanOpenInventory() const { return IsAlive() && !Battle; }
		bool CanOpenParty() const { return IsAlive() && !Battle; }
		bool CanTeleport() const { return IsAlive() && !Battle; }
		bool CanBattle() const { return !Battle && IsAlive() && Status == STATUS_NONE && !Invisible; }
		bool CanPVP() const { return !Battle && IsAlive(); }

		// Battle
		bool IsZoneOnCooldown(uint32_t Zone) { return BossCooldowns.find(Zone) != BossCooldowns.end(); }
		void GenerateNextBattle();
		int GenerateDamage();
		float GetAverageDamage() const { return (Attributes.at("MinDamage").Int + Attributes.at("MaxDamage").Int) / 2.0f; }
		float GetDamagePowerMultiplier(uint32_t DamageTypeID);

		// Actions
		ScopeType GetScope() { return Battle ? ScopeType::BATTLE : ScopeType::WORLD; }
		void RefreshActionBarCount();
		bool GetActionFromActionBar(_Action &ReturnAction, size_t Slot);
		void GetSummonsFromBuffs(std::vector<std::pair<_Summon, _StatusEffect *> > &Summons);
		bool CanEquipSkill(const _Item *Skill);

		// Skills
		bool HasLearned(const _Item *Skill) const;
		int GetSkillPointsAvailable() const { return SkillPoints - SkillPointsUsed; }
		void AdjustSkillLevel(uint32_t SkillID, int Amount, bool SoftMax);
		void AdjustMaxSkillLevel(uint32_t SkillID, int Amount);
		void InitializeSkillLevels();

		// UI
		bool IsTrading() { return WaitingForTrade || TradePlayer; }
		bool CanTrade() const;
		void ResetUIState(bool ResetMenuState=true);
		bool AddStatusEffect(_StatusEffect *StatusEffect);
		void DeleteStatusEffects();
		uint8_t GetStatus();
		void UpdateStatusTexture();

		// Unlocks
		void ClearUnlocks();
		int UnlockBySearch(const std::string &Search, int Count);
		bool HasUnlocked(const _Item *Item) const;

		// Base
		_Object *Object{nullptr};
		uint32_t CharacterID{0};
		uint32_t BuildID{1};
		double UpdateTimer{0.0};

		// Pointers
		_Battle *Battle{nullptr};
		_HUD *HUD{nullptr};

		// Render
		const ae::_Texture *StatusTexture{nullptr};
		const ae::_Texture *Portrait{nullptr};
		uint32_t PortraitID{0};

		// State
		std::unordered_map<uint32_t, double> BossCooldowns;
		std::unordered_map<uint32_t, int> BossKills;
		std::string PartyName;
		double IdleTime{0.0};
		int NextBattle{0};
		int Invisible{0};
		bool Hardcore{false};
		bool Offline{false};
		uint8_t Status{0};

		// Levels
		bool CalcLevelStats{true};
		int Level{0};
		int64_t ExperienceNeeded{0};
		int64_t ExperienceNextLevel{0};

		// Base attributes
		int64_t BaseMaxHealth{0};
		int64_t BaseMaxMana{0};
		int BaseMinDamage{0};
		int BaseMaxDamage{0};
		int BaseArmor{0};
		int BaseDamageBlock{0};
		int BaseBattleSpeed{100};
		int BaseSpellDamage{100};
		double BaseAttackPeriod{BATTLE_DEFAULTATTACKPERIOD};

		// Final attributes
		std::unordered_map<std::string, _Value> Attributes;
		std::unordered_map<std::string, int> BaseResistances;
		std::unordered_map<uint32_t, _SetData> Sets;
		int BattleSpeedBeforeBuffs;

		// Status effects
		std::vector<_StatusEffect *> StatusEffects;

		// Unlocks
		std::unordered_map<uint32_t, _Unlock> Unlocks;

		// Skills
		std::unordered_map<uint32_t, int> Skills;
		std::unordered_map<uint32_t, int> MaxSkillLevels;
		std::unordered_map<uint32_t, _Cooldown> Cooldowns;
		int SkillPoints{0};
		int SkillPointsUnlocked{0};
		int SkillPointsUsed{0};
		int SkillPointsOnActionBar{0};

		// Action bar
		std::vector<_Action> ActionBar;
		int BeltSize{ACTIONBAR_DEFAULT_BELTSIZE};
		int SkillBarSize{ACTIONBAR_DEFAULT_SKILLBARSIZE};

		// Events
		const _Vendor *Vendor{nullptr};
		const _Trader *Trader{nullptr};
		const _Blacksmith *Blacksmith{nullptr};
		const _Enchanter *Enchanter{nullptr};
		const _MinigameType *Minigame{nullptr};
		uint32_t Seed{0};
		uint32_t PaidSeed{0};

		// Trading
		_Object *TradePlayer{nullptr};
		int64_t TradeGold{0};
		bool WaitingForTrade{false};
		bool TradeAccepted{false};

		// Actions
		std::vector<_Object *> Targets;
		_Action Action;

		// Map
		ae::NetworkIDType LoadMapID{0};
		ae::NetworkIDType SpawnMapID{1};
		uint32_t SpawnPoint{0};
		double TeleportTime{0.0};

		// HUD
		bool MenuOpen{false};
		bool InventoryOpen{false};
		bool SkillsOpen{false};

		// Bots
		bool Bot{false};
		std::list<void *> Path;

	private:

		void CalculateStatBonuses(_StatChange &StatChange);

};
