/******************************************************************************
* choria
* Copyright (c) 2023 Alan Witkowski
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/
#pragma once

// Libraries
#include <ae/baseobject.h>
#include <ae/log.h>
#include <glm/vec2.hpp>
#include <glm/vec4.hpp>
#include <unordered_map>
#include <list>
#include <vector>
#include <string>
#include <cstdint>

// Forward Declarations
class _Character;
class _Inventory;
class _Fighter;
class _Controller;
class _Monster;
class _Map;
class _Battle;
class _Buff;
class _Stats;
class _Server;
class _Scripting;
class _StatChange;
class _StatusEffect;
class _HUD;
struct _Tile;
struct _ActionResult;
struct _Slot;

namespace ae {
	template<class T> class _Manager;
	class _Buffer;
	class _Texture;
	class _Element;
	class _Peer;
}

namespace Json {
	class Value;
}

// Classes
class _Object : public ae::_BaseObject {

	public:

		enum MoveDirectionType {
			MOVE_UP    = (1 << 0),
			MOVE_DOWN  = (1 << 1),
			MOVE_LEFT  = (1 << 2),
			MOVE_RIGHT = (1 << 3),
		};

		~_Object() override;

		// Components
		void CreateComponents();

		// Updates
		void Update(double FrameTime) override;
		void UpdateBot(double FrameTime);
		void Render(glm::vec4 &ViewBounds, const _Object *ClientPlayer=nullptr);
		void RenderBattle(_Object *ClientPlayer, double Time, bool ShowLevel);

		// Save
		void SerializeSaveData(Json::Value &Data) const;
		void UnserializeSaveData(const std::string &JsonString);

		// Network
		void SerializeCreate(ae::_Buffer &Data);
		void SerializeUpdate(ae::_Buffer &Data);
		void SerializeStats(ae::_Buffer &Data);
		void SerializeBattle(ae::_Buffer &Data);
		void SerializeStatusEffects(ae::_Buffer &Data);
		void SerializeBossCooldowns(ae::_Buffer &Data);
		void UnserializeCreate(ae::_Buffer &Data);
		void UnserializeStats(ae::_Buffer &Data);
		void UnserializeBattle(ae::_Buffer &Data, bool IsSelf);
		void UnserializeStatusEffects(ae::_Buffer &Data);
		void UnserializeBossCooldowns(ae::_Buffer &Data);
		void SendPacket(ae::_Buffer &Packet, bool Broadcast=true);
		void SendActionClear();

		// Logging
		void SetLogging(bool Value);

		// Stats
		bool IsMonster() const;
		_StatusEffect *UpdateStats(_StatChange &StatChange, _Object *Source=nullptr);
		void ApplyDeathPenalty(bool InBattle, double Penalty, int64_t BountyLoss);

		// Battles
		void UpdateMonsterAI(double FrameTime);
		void CreateBattleElement(ae::_Element *Parent);
		void RemoveBattleElement();
		void StopBattle();

		// Status effects
		void ResolveBuff(_StatusEffect *StatusEffect, const std::string &Function);

		// Actions
		bool SetActionUsing(ae::_Buffer &Data, ae::_Manager<_Object> *ObjectManager);
		bool CanInteractWith(const _Object *Object, int LevelRange, bool &HitLevelRestriction) const;

		// Movement
		void GetDirectionFromInput(int InputState, glm::ivec2 &Direction);
		int Move();

		// Trader
		void AcceptTrader(std::vector<_Slot> &Slots);

		// Minigames
		void SendSeed();

		// Map
		bool CanRespec() const;
		const _Tile *GetTile() const;
		ae::NetworkIDType GetMapID() const;

		// Path finding
		bool Pathfind(const glm::ivec2 &StartPosition, const glm::ivec2 &EndPosition);
		int GetInputStateFromPath();

		// Base
		std::string Name;
		ae::_LogFile Log;
		bool Logging{false};
		bool Muted{false};

		// Components
		_Character *Character{nullptr};
		_Inventory *Inventory{nullptr};
		_Fighter *Fighter{nullptr};
		_Controller *Controller{nullptr};
		_Monster *Monster{nullptr};

		// Pointers
		const _Stats *Stats{nullptr};
		_Map *Map{nullptr};
		_Scripting *Scripting{nullptr};
		_Server *Server{nullptr};
		ae::_Peer *Peer{nullptr};
		uint32_t QueuedMapChange{0};

		// Movement
		glm::ivec2 Position{0};
		glm::ivec2 ServerPosition{0};

		// Render
		const ae::_Texture *ModelTexture{nullptr};
		uint32_t ModelID{0};
		uint32_t BossZoneID{0};
		int Light{0};

		// Compression
		uint8_t UpdateID{0};
		bool Changed{false};
		glm::ivec2 OldPosition{0};
		int OldStatus{0};
		int OldInvisible{0};
		int64_t OldBounty{0};
		int OldLight{0};

	private:

};

inline bool CompareObjects(const _Object *First, const _Object *Second) {
	return First->NetworkID < Second->NetworkID;
}
